// qtre.cpp
// Revision 06-feb-2001

#include <iostream>
#include <iomanip>
#include <unistd.h>
#include <stdexcept>

#include "termstream.h"
#include "text.h"
#include "qperl.h"
#include "menu.h"

#include <assert.h>
#define ASSERT assert

const IdMenu
	ID_OPEN= 1, ID_SAVE= 2, ID_QUIT= 3, ID_HELP= 4,
	ID_SELECT= 5, ID_CUT= 6, ID_COPY= 7, ID_PASTE= 8;

class StatusLine {
	otermstream ot;
	size_t max;
public:
	StatusLine (int ancho, int y);
	void settext (std::string str);
	void lincol (int y, int x);
};

StatusLine::StatusLine (int ancho, int y) :
	ot (1, ancho, y, 0)
{
	ot.inverse (true);
	max= ancho - 11;
	ot << at (0, 0) << std::string (ancho, ' ') << flush;
}

void StatusLine::settext (std::string str)
{
	size_t l= str.size ();
	if (l < max)
		str+= std::string (max - l, ' ');
	else if (l > max)
		str= str.substr (0, max);
	ot << at (0, 11) << ' ' << str << flush;
}

void StatusLine::lincol (int y, int x)
{
	ot << at (0, 0) << setw (5) << y << ' ' <<
		setw (5) << x <<
		flush;
}

class Screen {
	Text & text;
	Text * clipboard;
	StatusLine & status;
	Menu & menu;
	otermstream ot;
	bool fExit;
	Selection sel;
	size_t maxlin, maxcol;
	size_t lin, col;
	size_t inilin, inicol;
	static int perl_link (void * thisclass,
		qpFuncCode f, int n, void *data);
	bool fRedrawAll;
	bool fSelecting;
public:
	Screen (Text & ntext, StatusLine & st, Menu & mn,
		int alto, int ancho, int y, int x);
	void beginselect ();
	void endselect ();
	void copyselection ();
	void deleteselection ();
	void cut ();
	void copy ();
	void paste ();
	void gotoline (size_t l);
	char getcurchar ();
	void getline (int nline, char * str);
	void instext (const char * str);
	void putline (size_t i);
	void putall ();
	void scrollup ();
	void scrolldown ();
	void adjustmaxcol ();
	void movesel (const PosText & opt, const PosText & npt);
	void home ();
	void end ();
	bool up ();
	void pageup ();
	bool down ();
	void pagedown ();
	bool back ();
	void forward ();
	void deletechar ();
	void backdelete ();
	void inschar (unsigned char c);
	void sendchar (unsigned char c);
	void adjustinicol ();
	void adjustinilin ();
	void save ();
	void help ();
	void menuselect ();
	void sendkey (int c);
	void deleteline ();
	void replaceline (char * str);
	void interact ();
};

int Screen::perl_link (void *thisclass, qpFuncCode f, int n, void * data)
{
	Screen *p= reinterpret_cast <Screen *> (thisclass);
	switch (f) {
	case qpUp:
		p->up ();
		break;
	case qpDown:
		p->down ();
		break;
	case qpLeft:
		p->back ();
		break;
	case qpRight:
		p->forward ();
		break;
	case qpHome:
		p->home ();
		break;
	case qpEnd:
		p->end ();
		break;
	case qpGotoLine:
		p->gotoline (n);
		break;
	case qpGetChar:
		return p->getcurchar ();
		break;
	case qpDelChar:
		p->deletechar ();
		break;
	case qpDelLine:
		p->deleteline ();
		break;
	case qpReplaceLine:
		p->replaceline ( (char *) data);
		break;
	case qpInsText:
		p->instext ( (char *) data);
		break;
	case qpLine:
		return p->lin;
	case qpNumlines:
		return p->text.lines ();
	case qpCharLine:
		return p->text [n].size ();
	case qpGetLine:
		p->getline (n, (char *) data);
		break;
	case qpStatus:
		p->status.settext ( (char *) data);
		break;
	}
	return 0;
}

void Screen::getline (int nline, char * str)
{
	TextLine & s= text [nline];
	size_t l= s.size ();
	size_t i;
	for (i= 0; i < l; ++i)
		str [i]= s [i];
	str [i]= '\0';
}

void Screen::instext (const char * str)
{
	for (; * str != '\0'; ++str)
		sendchar (* str);
}

Screen::Screen (Text & ntext, StatusLine & st, Menu & mn,
	int alto, int ancho, int y, int x) :
	text (ntext),
	status (st),
	menu (mn),
	ot (alto, ancho, y, x)
{
	inilin= 0;
	inicol= 0;
	maxlin= alto - 1; maxcol= ancho - 1;
	lin= 0; col= 0;
	fSelecting= false;
	clipboard= NULL;
	//ot.bgcolor (3);
	fExit= false;
}

void Screen::beginselect ()
{
	bool fSelPrev= sel.begin () != sel.end ();
	TextLine & s= text [lin];
	size_t pos= s.col2pos (col);
	sel.init (lin, pos);
	fSelecting= true;
	if (fSelPrev) fRedrawAll= true;
}

void Screen::endselect ()
{
	fSelecting= false;
}

void Screen::copyselection ()
{
	endselect ();
	if (sel.begin () == sel.end () )
		return;
	if (clipboard) delete clipboard;
	clipboard= new Text;
	size_t linfirst= sel.begin ().lin;
	size_t linlast= sel.end ().lin;
	size_t posfirst= sel.begin ().pos;
	size_t poslast= sel.end ().pos;
	std::string s;
	if (linfirst == linlast) {
		s= text [linfirst].substr (posfirst, poslast - posfirst);
		clipboard->operator [] (0)= s;
	}
	else {
		s= text [linfirst].substr (posfirst);
		clipboard->operator [] (0)= s;
		for (size_t i= linfirst + 1; i < linlast; ++i)
			clipboard->push_back (text [i].str () );
		s= text [linlast].substr (0, poslast);
		clipboard->push_back (s);
	}
}

void Screen::deleteselection ()
{
	endselect ();
	if (sel.begin () == sel.end () )
		return;
	size_t linbeg= sel.begin ().lin;
	size_t linend= sel.end ().lin;
	if (linbeg == linend)
		text [linbeg]=
			text [linbeg].substr (0, sel.begin ().pos) +
			text [linbeg].substr (sel.end ().pos);
	else {
		text [linbeg]= text [linbeg].substr (0, sel.begin ().pos);
		text [linend]= text [linend].substr (sel.end ().pos);
		for (size_t lin= linbeg + 1; lin < linend; ++lin)
			text.deleteline (linbeg + 1);
		text.joinline (linbeg);
	}
	lin= linbeg;
	col= text [lin].pos2col (sel.begin ().pos);
	sel.init (0, 0);
	adjustinilin ();
	fRedrawAll= true;
}

void Screen::cut ()
{
	endselect ();
	if (sel.begin () == sel.end () )
		return;
	status.settext ("Cortando");
	copyselection ();
	deleteselection ();
}

void Screen::copy ()
{
	endselect ();
	if (sel.begin () == sel.end () )
		return;
	status.settext ("Copiando");
	copyselection ();	
}

void Screen::paste ()
{
	endselect ();
	if (! clipboard)
		return;
	status.settext ("Pegando");
	size_t l= clipboard->lines ();
	for (size_t i= 0; i < l; ++i) {
		TextLine & s= clipboard->operator [] (i);
		for (size_t j= 0; j < s.size (); ++j)
			sendchar (s [j]);
		if (i != l - 1)
			sendchar ('\n');			
	}
}

void Screen::gotoline (size_t l)
{
	PosText pt;
	if (fSelecting) pt (lin, text [lin].col2pos (col) );
	size_t iniant= inilin;
	if (l < inilin) {
		lin= l;
		inilin= l;
	}
	else
		if (l < inilin + maxlin)
			lin= l;
		else {
			inilin= l - maxlin;
			lin= l;
		}
	col= 0;
	if (fSelecting)
		movesel (pt, PosText (lin, text [lin].col2pos (col) ) );
	if (inilin != iniant || fSelecting)
		fRedrawAll= true;
	adjustinicol ();
}

char Screen::getcurchar ()
{
	TextLine & s= text [lin];
	size_t pos= s.col2pos (col);
	if (pos >= s.size () ) {
		if (lin >= text.lines () - 1) return '\0';
		else return '\n';
	}
	return s [pos];
}

void Screen::putline (size_t l)
{
	ASSERT (l <= maxlin);
	if (fRedrawAll)
		return;
	size_t lin= l + inilin;
	ot << at (l, 0);
	if (lin >= text.lines () ) {
		ot << clreol;
	}
	else {
		size_t posBegSel, posEndSel;
		bool fSelected= lin >= sel.begin ().lin &&
			lin <= sel.end ().lin && sel.begin () != sel.end ();
		bool fSelectedBeg= lin == sel.begin ().lin;
		if (fSelectedBeg)
			posBegSel= sel.begin ().pos;
		bool fSelectedEnd= lin == sel.end ().lin;
		if (fSelectedEnd)
			posEndSel= sel.end ().pos;
		if (fSelected && ! fSelectedBeg)
			ot << inverse (true);
		TextLine & s= text [lin];
		size_t lim= s.size ();
		size_t pos= s.col2pos (inicol);
		size_t ini;
		if (pos <= lim) {
			ini= s.pos2col (pos);
			ASSERT (ini == inicol || ini == inicol + 1);
		}
		else ini= 0;
		size_t acol;
		unsigned char c;
		if ( (ini & 1) && pos <= lim) {
			ASSERT (ini == inicol + 1);
			ASSERT (pos > 0);
			ASSERT (Text::escapable (s [pos - 1] ) );
			ot << Text::escaped (s [pos - 1] );
		}
		for (acol= 0; acol <= maxcol && pos < lim; ++pos)
		{
			if (fSelectedBeg && pos >= posBegSel) {
				ot << inverse (true);
				fSelectedBeg= false;
			}
			if (fSelectedEnd && pos >= posEndSel) {
				ot << inverse (false);
				fSelectedEnd= false;
			}
			c= s [pos];
			if (c == '\t') {
				acol= Text::nexttab (acol);
				ot << '\t';
			}
			else
				if (Text::escapable (c) ) {
					acol+= 2;
					ot << '^';
					if (acol <= maxcol + 1)
						ot << Text::escaped (c);
				}
				else {
					++acol;
					// Prueba
					//if (c == 155) c= '';
					ot << c;
				}
		}
		if (acol <= maxcol)
			ot << clreol;
		ot << flush << inverse (false);
	}
}

void Screen::putall ()
{
	fRedrawAll= false;
	for (size_t i= 0; i <= maxlin; ++i)
		putline (i);
}

void Screen::scrollup ()
{
	--inilin;
	ot << at (0, 0) << insline;
	putline (0);
}

void Screen::scrolldown ()
{
	++inilin;
	ot << at (0, 0) << delline;
	putline (maxlin);
}

void Screen::adjustmaxcol ()
{
	TextLine & s= text [lin];
	size_t m= s.colmax ();
	if (col > m)
		col= m;
	else col= s.pos2col (s.col2pos (col) );
}

void Screen::movesel (const PosText &opt, const PosText &npt)
{
	if (opt == sel.end () )
		sel.moveend (npt);
	else {
		ASSERT (opt == sel.begin () );
		sel.movebegin (npt);
	}
}

void Screen::home ()
{
	if (col == 0)
		return;
	PosText opt;
	if (fSelecting) opt (lin, text [lin].col2pos (col) );
	col= 0;
	adjustinicol ();
	if (fSelecting) {
		movesel (opt, PosText (lin, 0) );
		putline (lin - inilin);
	}
}

void Screen::end ()
{
	TextLine & s= text [lin];
	size_t npos= s.size ();
	size_t acol= col;
	col= s.pos2col (npos);
	if (col == acol)
		return;
	adjustinicol ();
	if (fSelecting) {
		movesel (PosText (lin, s.col2pos (acol) ),
			PosText (lin, npos) );
		putline (lin - inilin);
	}
}

bool Screen::up ()
{
	if (lin == 0)
		return false;
	PosText opt;
	if (fSelecting) opt (lin, text [lin].col2pos (col) );
	--lin;
	adjustmaxcol ();
	if (fSelecting) {
		size_t npos= text [lin].col2pos (col);
		movesel (opt, PosText (lin, npos) );
		putline (opt.lin - inilin);
	}
	if (lin < inilin) {
		scrollup ();
	}
	else if (fSelecting)
		putline (lin - inilin);
	return true;
}

void Screen::pageup ()
{
	size_t oini= inilin;
	PosText opt;
	if (fSelecting) opt (lin, text [lin].col2pos (col) );
	if (inilin > 0) {
		if (inilin > maxlin) inilin-= maxlin;
		else inilin= 0;
		if (lin > maxlin) lin-= maxlin;
		else lin= 0;
	}
	else {
		if (lin == 0) return;
		lin= 0;
	}
	adjustmaxcol ();
	if (fSelecting)
		movesel (opt, PosText (lin, text [lin].col2pos (col) ) );
	if (oini != inilin || fSelecting)
		putall ();
}

bool Screen::down ()
{
	if (lin + 1 >= text.lines () ) return false;
	PosText opt;
	if (fSelecting) opt (lin, text [lin].col2pos (col) );
	++lin;
	adjustmaxcol ();
	if (fSelecting) {
		size_t npos= text [lin].col2pos (col);
		movesel (opt, PosText (lin, npos) );
		putline (opt.lin - inilin);
	}
	if (lin > inilin + maxlin) {
		scrolldown ();
		ASSERT (lin == inilin + maxlin);
	}
	else if (fSelecting)
		putline (lin - inilin);
	return true;
}

void Screen::pagedown ()
{
	if (inilin >= text.lines () - 1) {
		ASSERT (inilin == text.lines () - 1);
		return;
	}
	PosText opt;
	if (fSelecting) opt (lin, text [lin].col2pos (col) );
	if (inilin + maxlin < text.lines () ) {
		inilin+= maxlin;
		lin+= maxlin;
		if (lin >= text.lines () )
			lin= text.lines () - 1;
	}
	else {
		inilin= text.lines () - 1;
		lin= inilin;
	}
	adjustmaxcol ();
	if (fSelecting)
		movesel (opt, PosText (lin, text [lin].col2pos (col) ) );
	putall ();
}

bool Screen::back ()
{
	TextLine & s= text [lin];
	size_t pos= s.col2pos (col);
	PosText opt;
	if (fSelecting) opt (lin, pos);
	if (pos > 0) {
		--pos;
		col= s.pos2col (pos);
	}
	else {
		if (lin == 0) return false;
		--lin;
		col= text [lin].colmax ();
		pos= text [lin].size ();
	}
	if (fSelecting)
		movesel (opt, PosText (lin, pos) );
	if (lin < inilin)
		scrollup ();
	else
		if (fSelecting && lin != opt.lin)
			putline (opt.lin - inilin);
	if (fSelecting)
		putline (lin - inilin);
	return true;
}

void Screen::forward ()
{
	TextLine & s= text [lin];
	size_t pos= s.col2pos (col);
	PosText opt;
	if (fSelecting) opt (lin, pos);
	if (pos < s.size () ) {
		++pos;
		col= s.pos2col (pos);
	}
	else {
		if (lin + 1 >= text.lines () )
			return;
		++lin;
		col= 0; pos= 0;
	}
	if (fSelecting)
		movesel (opt, PosText (lin, pos) );
	if (lin - inilin > maxlin)
		scrolldown ();
	else
		if (fSelecting && lin != opt.lin)
			putline (opt.lin - inilin);
	if (fSelecting)
		putline (lin - inilin);
}

void Screen::deletechar ()
{
	TextLine & s= text [lin];
	size_t pos= s.col2pos (col);

	if (pos >= s.size () ) {
		if (lin < text.lines () - 1) {
			text.joinline (lin);
			ot << clrscr;
			putall ();
		}
	}
	else {
		s.deletechar (col);
		putline (lin - inilin);
	}
}

void Screen::backdelete ()
{
	if (back () )
		deletechar ();
}

void Screen::inschar (unsigned char c)
{
	ASSERT (c != 0);
	text [lin].insertchar (col, c);
}

void Screen::sendchar (unsigned char c)
{
	switch (c) {
	case '\n':
		text.insertline (lin, col);
		down ();
		ASSERT (lin > 0);
		putline (lin - inilin - 1);
		ot << at (lin - inilin, 0) << insline;
		putline (lin - inilin);
		col= 0;
		break;
	default:
		if (c >= 32 || c == '\t') {
			inschar (c);
			forward ();
			putline (lin - inilin);
		}
	}
}

void Screen::adjustinicol ()
{
	if (col < inicol)
		inicol= (col / 8) * 8;
	else
		if (col > inicol + maxcol)
			inicol= ( (col - maxcol + 7) / 8) * 8;
		else return;
	fRedrawAll= true;
}

void Screen::adjustinilin ()
{
	if (lin < inilin)
		inilin= lin;
	else
		if (lin > inilin + maxlin)
			inilin= lin - maxlin;
		else return;
	fRedrawAll= true;
}

void Screen::save ()
{
	text.save ();
}

void Screen::help ()
{
	status.settext ("F2: menu F3: guardar F8: salir, "
		"no hay mas ayuda por ahora");
}

void Screen::menuselect ()
{
	ot << cursor (false);
	IdMenu id;
	if (menu.select (& id) ) {
		switch (id) {
		case ID_OPEN:
			status.settext ("Opcion no disponible");
			break;
		case ID_SAVE:
			save ();
			break;
		case ID_QUIT:
			fExit= true;
			break;
		case ID_HELP:
			help ();
			break;
		case ID_SELECT:
			if (! fSelecting)
				beginselect ();
			else
				endselect ();
			break;
		case ID_CUT:
			cut ();
			break;
		case ID_COPY:
			copy ();
			break;
		case ID_PASTE:
			paste ();
			break;
		default:
			ASSERT (false);
		}
	}
	ot << cursor (true);
}

void Screen::sendkey (int c)
{
	using namespace key;
	int c2;
	switch (c) {
	case Home:
		home ();
		break;
	case End:
		end ();
		break;
	case Up:
		up ();
		break;
	case Down:
		down ();
		break;
	case Left:
		back ();
		break;
	case Right:
		forward ();
		break;
	case PageUp:
		pageup ();
		break;
	case PageDown:
		pagedown ();
		break;
	case F1:
		help ();
		break;
	case F2:
		menuselect ();
		break;
	case F3:
		save ();
		break;
	case F7:
		putall ();
		break;
	case F8:
		fExit= true;
		break;
	case Backspace:
		backdelete ();
		break;
	case Delete:
		deletechar ();
		break;
	case '\x03': // Control-c, copy
		copy ();
		break;
	case '\x08': // Control-h
		backdelete ();
		break;
	case '\x0B': // Control-k, macro.
		c2= ot.getkey ();
		if ( (c2 >= '0' && c2 <= '9') || (c2 >= 'a' && c2 <= 'z') ) 
			if (qperl_macro (c2, perl_link, this) )
				status.settext ("Macro no definida");
		break;
	case '\x0C': // Control-l, prueba.
		gotoline (40);
		break;
	case '\x0E': // Control-n, menu.
		menuselect ();
		break;
	case '\x11': // Control-q, quit.
		fExit= true;
		break;
	case '\x13': // Control-s, select.
		if (! fSelecting)
			beginselect ();
		else
			endselect ();
		break;
	case '\x16': // Control-v, paste
		paste ();
		break;
	case '\x18': // Control-x, cut
		cut ();
		break;
	case '\x19': // Control-y, borra linea
		deleteline ();
		break;
	default:
		if (c <= 255) {
			sendchar (c);
		}
	}
}

void Screen::deleteline ()
{
	text.deleteline (lin);
	ot << at (lin - inilin,0) << delline;
	putline (maxlin);
	if (lin >= text.lines () )
		up ();
	else
		adjustmaxcol ();
}

void Screen::replaceline (char *str)
{
	status.settext ("Replace line");
	text [lin]= std::string (str);
	adjustmaxcol ();
	putline (lin - inilin);
}

void Screen::interact ()
{
	int c;
	do {
		fRedrawAll= false;
		status.lincol (lin, col);
		ot << at (lin - inilin, col - inicol);
		c= ot.getkey ();
		sendkey (c);
		adjustinicol ();
		if (fRedrawAll)
			putall ();
	} while (! fExit);
}

void bloque_principal (int argc, char * * argv)
{
	otermstream pan;
	Text *ptext= NULL;
	try {
		pan.setcolor (1, Color::Cyan, Color::Black);
		int lin= pan.lines ();
		int col= pan.columns ();
		otermstream otmenubar (1, col, 0, 0);
		StatusLine stat (col, lin - 1);
		otmenubar.color (1);
		Menu menuFile (otmenubar);
		menuFile.addOption (Option ("Abrir",    ID_OPEN) );
		menuFile.addOption (Option ("Guardar",  ID_SAVE) );
		menuFile.addOption (Option ("Terminar", ID_QUIT) );
		Menu menuEdit (otmenubar);
		menuEdit.addOption (Option ("Seleccionar", ID_SELECT) );
		menuEdit.addOption (Option ("Cortar",      ID_CUT) );
		menuEdit.addOption (Option ("Copiar",      ID_COPY) );
		menuEdit.addOption (Option ("Pegar",       ID_PASTE) );
		Menu menu (otmenubar);
		menu.addOption (Option ("Archivo", & menuFile) );
		menu.addOption (Option ("Edicion", & menuEdit) );
		menu.addOption (Option ("Ayuda", ID_HELP) );
		menu.show ();
		stat.settext ("Estado");

		if (argc > 1) ptext= new Text (argv [1]);
		else ptext= new Text;
		Screen screen (* ptext, stat, menu, lin - 2, col, 1, 0);
		screen.putall ();
		screen.interact ();
		delete ptext;
	}
	catch (...) {
		if (ptext) delete ptext;
		throw;
	}
}

int main (int argc, char * * argv, char * * env)
{
	using std::cout;
	using std::endl;
	qperl_init (env);
	try {
		bloque_principal (argc, argv);
	}
	catch (std::exception &e) {
		cout << endl << e.what ();
	}
	catch (...) {
		cout << endl << "Error inesperado";
	}
	qperl_destroy ();
	cout << endl;
	return 0;
} // main

// Fin de qtre.cpp
