// termstream.h

#ifndef _TERMSTREAM_H
#define _TERMSTREAM_H 1

#include <iostream>
#include <streambuf.h>

class Color {
public:
	typedef short value;
	static const value
		Black= 0,
		Red= 1,
		Green= 2,
		Yellow= 3,
		Blue= 4,
		Magenta= 5,
		Cyan= 6,
		White= 7;
};

class termbuf : public streambuf
{
public:
	explicit termbuf (ios::openmode which= ios::in | ios::out)
		: streambuf (which)
	{
		setp (defbuf, defbuf + bufsize);
	}
	void setWindow (void *window) { w= window; }
	int getkey ();
	virtual int sync ();
protected:
	virtual int overflow (int ch= EOF);
	virtual int underflow ();
private:
	static const streamsize bufsize= 64;
	char defbuf [bufsize];
	void *w;
	class Control {
	public:
		Control ();
		~Control ();
	private:
		static int n;
	};
	Control control;
};

class termstreambase : virtual public ios {
protected:
	termbuf __my_tb;
public:
	termbuf * rdbuf ()
		{ return & __my_tb; }
	inline void sync () { __my_tb.sync (); }
	int getkey () { return __my_tb.getkey (); }
	int lines ();
	int columns ();
protected:
	termstreambase (ios::openmode which)
		: __my_tb (which)
	{
		init (&__my_tb);
	}
	void setWindow (void *window);
	void *w;
};

class ts_bool {
public:
	ostream & (* f) (ostream &, bool);
	bool b;
	ts_bool (ostream & (*ff) (ostream &, bool), bool bb)
		: f (ff), b (bb) {}
};

class ts_2int {
public:
	ostream & (* f) (ostream &, int, int);
	int i1, i2;
	ts_2int (ostream & (*ff) (ostream &, int, int), int ii1, int ii2)
		: f (ff), i1 (ii1), i2 (ii2) {}
};

class ts_short {
public:
	ostream & (* f) (ostream &, short);
	short s;
	ts_short (ostream & (*ff) (ostream &, short), short ss)
		: f (ff), s (ss) {}
};

class otermstream : public termstreambase, public ostream {
public:
	otermstream (ios::openmode which= ios::out);
	otermstream (int nlines, int ncols, int y, int x);
	void clrscr ();
	void clreol ();
	void insertline ();
	void deleteline ();
	void at (int y, int x);
	inline void gotoxy (int x, int y) { at (y, x); }
	void inverse (bool flag);
	void delay (bool flag);
	void cursor (bool flag);
	void setcolor (short n, Color::value f, Color::value b);
	void color (short c);
	friend ostream & operator << (ostream &os, const ts_bool &tb)
		{ return tb.f (os, tb.b); }
	friend ostream & operator << (ostream &os, const ts_2int &t2i)
		{ return t2i.f (os, t2i.i1, t2i.i2); }
	friend ostream & operator << (ostream &os, const ts_short &ts)
		{ return ts.f (os, ts.s); }
private:
	void setback ();
};

inline ostream & inverse (ostream &os, bool b)
{
	otermstream & o= dynamic_cast<otermstream &> (os);
	o.inverse (b);
	return os;
}

inline ostream & delay (ostream &os, bool b)
{
	otermstream & o= dynamic_cast<otermstream &> (os);
	o.delay (b);
	return os;
}

inline ostream & cursor (ostream &os, bool b)
{
	otermstream & o= dynamic_cast<otermstream &> (os);
	o.cursor (b);
	return os;
}

inline ostream & gotoxy (ostream &os, int x, int y)
{
	otermstream & o= dynamic_cast<otermstream &> (os);
	o.gotoxy (x, y);
	return os;
}

inline ostream & at (ostream &os, int y, int x)
{
	otermstream & o= dynamic_cast<otermstream &> (os);
	o.at (y, x);
	return os;
}

inline ostream & color (ostream &os, short c)
{
	otermstream & o= dynamic_cast<otermstream &> (os);
	o.color (c);
	return os;
}


inline ts_bool inverse (bool b) { return ts_bool (inverse, b); }
inline ts_bool delay (bool b) { return ts_bool (delay, b); }
inline ts_bool cursor (bool b) { return ts_bool (cursor, b); }
inline ts_2int gotoxy (int x, int y) { return ts_2int (gotoxy, x, y); }
inline ts_2int at (int y, int x) { return ts_2int (at, y, x); }
inline ts_short color (short c) { return ts_short (color, c); }

inline ostream & clrscr (ostream &os)
{
	otermstream & o= dynamic_cast<otermstream &> (os);
	o.clrscr ();
	return os;
}

inline ostream & clreol (ostream &os)
{
	otermstream & o = dynamic_cast<otermstream &> (os);
	o.clreol ();
	return os;
}

inline ostream & delline (ostream &os)
{
	otermstream & o= dynamic_cast<otermstream &> (os);
	o.deleteline ();
	return os;
}

inline ostream & insline (ostream &os)
{
	otermstream & o= dynamic_cast<otermstream &> (os);
	o.insertline ();
	return os;
}

namespace key {
const int
	Left=      0x0100,
	Right=     0x0101,
	Up=        0x0102,
	Down=      0x0103,
	PageUp=    0x0104,
	PageDown=  0x0105,
	Backspace= 0x0106,
	Delete=    0x0107,
	Home=      0x0108,
	End=       0x0109,
	F1=  0x0201,
	F2=  0x0202,
	F3=  0x0203,
	F4=  0x0204,
	F5=  0x0205,
	F6=  0x0206,
	F7=  0x0207,
	F8=  0x0208,
	F9=  0x0209,
	F10= 0x020A,
	F11= 0x020B,
	F12= 0x020C;
};

#endif
// Fin de termstream.h

