// text.cpp

#include <string>
#include <fstream>
#include <vector>

#include "text.h"

const size_t size_max= size_t (0) - 1;

size_t TextLine::col2pos (size_t col)
{
	size_t pos= 0, i= 0;
	unsigned char c;
	size_t l= size ();
	while (i < col) {
		if (pos >= l)
			return size_max;
		c= s [pos];
		if (c == '\t')
			i= Text::nexttab (i);
		else
			if (Text::escapable (c) ) i+= 2;
			else ++i;
		++pos;
	}
	return pos;
}

size_t TextLine::pos2col (size_t pos)
{
	size_t col= 0;
	unsigned char c;
	size_t l= size ();
	for (size_t i= 0; i < pos; ++i) {
		if (i >= l)
			return size_max;
		c= s [i];
		if (c == '\t')
			col= Text::nexttab (col);
		else
			if (Text::escapable (c) ) col+= 2;
			else ++col;
	}
	return col;
}

size_t TextLine::colmax ()
{
	return pos2col (size () );
}

void TextLine::deletechar (size_t col)
{
	size_t pos= col2pos (col);
	s.replace (pos, 1, "");
}

void TextLine::insertchar (size_t col, char c)
{
	size_t pos= col2pos (col);
	if (s.size () <= pos) {
		if (s.size () < pos)
			s+= string (pos - s.size (), ' ');
		s+= c;
	}
	else s.insert (pos, 1, c);
}

ostream & operator << (ostream &os, const TextLine & tl)
{
	os << tl.s;
	return os;
}

Text::Text ()
{
	t.push_back ();
	nlines= t.size ();
}

Text::Text (const char *strFile)
{
	filename= strFile;
	ifstream in (strFile);
	in.unsetf (ios::skipws);
	char c;
	std::string s;
	t.push_back ();
	size_t l= 0;
	while (in >> c) {
		if (c == '\n') {
			t [l++]= s;
			t.push_back ();
			s= std::string ();
		}
		else
			s+= c;
	}
	if (s.size () > 0)
		t [l]= s;
	nlines= t.size ();
}

void Text::save ()
{
	if (filename.size () == 0)
		return; // Por ahora
	ofstream o (filename.c_str () );
	size_t lastline= nlines - 1;
	for (size_t i= 0; i <= lastline; ++i) {
		o << t [i];
		if (i < lastline)
			o << std::endl;
	}
	o.close ();
}

void Text::insertline (size_t lin, size_t col)
{
	t.insert (t.begin () + lin + 1, std::string("") );
	TextLine & ant= t [lin];
	TextLine & sigue= t [lin + 1];
	size_t pos= ant.col2pos (col);
	if (pos < ant.size () ) {
		sigue= ant.substr (pos);
		ant= ant.substr (0, pos);
	}
	++nlines;
}

void Text::deleteline (size_t lin)
{
	t.erase (t.begin () + lin);
	--nlines;
}

void Text::joinline (size_t lin)
{
	t [lin]+= t [lin + 1];
	t.erase (t.begin () + lin + 1);
	--nlines;
}

// Fin de text.cpp