// text.h

#include <string>
#include <vector>

class TextLine {
	std::string s;
public:
	TextLine () : s () {}
	TextLine (const std::string & snue) : s (snue) {}
	void operator = (std::string snue)
		{ s= snue; }
	void operator += (const TextLine & tnue)
		{ s+= tnue.s; }
	inline char operator [] (size_t n) { return s [n]; }
	size_t size () { return s.size (); }
	std::string str () { return s; }
	std::string substr (size_t n, size_t fin= std::string::npos)
		{ return s.substr (n, fin); }
	size_t col2pos (size_t col);
	size_t pos2col (size_t pos);
	size_t colmax ();
	void deletechar (size_t col);
	void insertchar (size_t col, char c);
	friend ostream & operator << (ostream & os, const TextLine & tl);
};

class PosText {
public:
	PosText () : lin (0), pos (0) { }
	PosText (size_t nlin, size_t npos) : lin (nlin), pos (npos) { }
	size_t lin, pos;
	bool operator < (const PosText &pt) const
	{
		if (lin < pt.lin)
			return true;
		if (lin == pt.lin && pos < pt.pos)
			return true;
		return false;
	}
	bool operator == (const PosText &pt) const
		{ return lin == pt.lin && pos == pt.pos; }
	void operator () (size_t nlin, size_t npos)
		{ lin= nlin; pos= npos; }
};

class Selection {
public:
	void init (int lin, int pos)
	{
		ptBegin (lin, pos);
		ptEnd (lin, pos);
	}
	const PosText & begin () { return ptBegin; }
	const PosText & end () { return ptEnd; }
	void movebegin (const PosText & npt) 
	{
		ptBegin= npt;
		adjust ();
	}
	void movebegin (int lin, int pos) { movebegin (PosText (lin, pos) ); }
	void moveend (const PosText & npt)
	{
		ptEnd= npt;
		adjust ();
	}
	void moveend (int lin, int pos) { moveend (PosText (lin, pos) ); }
private:
	PosText ptBegin, ptEnd;
	inline void adjust ()
		{ if (ptEnd < ptBegin) swap (ptBegin, ptEnd); }
};

class Text {
	std::string filename;
	std::vector<TextLine> t;
	size_t nlines;
	static const unsigned char EscAlt= (unsigned char) '\x9B';
public:
	Text ();
	Text (const char *strFile);
	static inline bool escapable (unsigned char c)
	{
		if (c < ' ' || c == '\x7F' || c == EscAlt) return true;
		return false;
	}
	static inline unsigned char escaped (unsigned char c)
	{
		if (c == '\x7F') return '?';
		if (c == EscAlt) return '';
		return c + '@';
	}
	static inline size_t nexttab (size_t col)
		//{ return ( (col + 8) / 8) * 8; }
		{ return (col / 8 + 1) * 8; }
	void save ();
	TextLine & operator [] (size_t n)
		{ return t [n]; }
	size_t lines () { return nlines; }
	void insertline (size_t lin, size_t col);
	void deleteline (size_t lin);
	void joinline (size_t lin);
	void push_back (const std::string & s)
		{ t.push_back (s); ++nlines; }
};

// Fin de text.h
